from clickhouse_query.functions import base


class Equals(base._Function2Args):
    function = "equals"


class NotEquals(base._Function2Args):
    function = "notEquals"


class Less(base._Function2Args):
    function = "less"


class Greater(base._Function2Args):
    function = "greater"


class LessOrEquals(base._Function2Args):
    function = "lessOrEquals"


class GreaterOrEquals(base._Function2Args):
    function = "greaterOrEquals"



from clickhouse_query import utils
from clickhouse_query.functions import base


class _Logical(base.Function):
    def __clickhouse_query_function_sql__(self, *, uid_generator):
        if len(self.args) + len(self.kwargs) == 0:
            raise ValueError("At least one argument is required")

        if len(self.args) + len(self.kwargs) == 1:
            return "", {}

        return super().__clickhouse_query_function_sql__(uid_generator=uid_generator)

    def __clickhouse_query_function_args_sqls__(self, *, uid_generator):
        function_args_sqls_list, function_args_params = super().__clickhouse_query_function_args_sqls__(
            uid_generator=uid_generator
        )
        for item, value in self.kwargs.items():
            expression = utils._get_expression(value)
            condition = utils._extract_condition(item, expression)
            sql, params = utils.get_sql(condition, uid_generator=uid_generator)
            function_args_sqls_list.append(sql)
            function_args_params.update(params)

        return function_args_sqls_list, function_args_params


class And(_Logical):
    function = "and"


class Or(_Logical):
    function = "or"


class Not(base._Function1Args):
    function = "not"


class Xor(_Logical):
    function = "xor"



from clickhouse_query.functions import base


class _AggregationFunctionWithMaxSizeParam(base.AggregationFunctionWithParams):
    def __init__(self, x, max_size=None):
        if max_size is None:
            super().__init__(x)
        else:
            super().__init__(x, params=[max_size])


class Count(base.AggregationFunction):
    function = "count"

    def __init__(self, arg=None):
        if arg is None:
            super().__init__()
        else:
            super().__init__(arg)


class Min(base._AggregationFunction1Args):
    function = "min"


class Max(base._AggregationFunction1Args):
    function = "max"


class Sum(base._AggregationFunction1Args):
    function = "sum"


class Avg(base._AggregationFunction1Args):
    function = "avg"


class Any(base._AggregationFunction1Args):
    function = "any"


class FirstValue(base._AggregationFunction1Args):
    function = "first_value"


class LastValue(base._AggregationFunction1Args):
    function = "last_value"


class ArgMin(base._AggregationFunction2Args):
    function = "argMin"


class ArgMax(base._AggregationFunction2Args):
    function = "argMax"


class AvgWeighted(base._AggregationFunction2Args):
    function = "avgWeighted"


class TopK(base.AggregationFunctionWithParams):
    function = "topK"

    def __init__(self, column, *, k):
        super().__init__(column, params=[k])


class TopKWeighted(base.AggregationFunctionWithParams):
    function = "topKWeighted"

    def __init__(self, column, weight, *, k):
        super().__init__(column, weight, params=[k])


class GroupArray(_AggregationFunctionWithMaxSizeParam):
    function = "groupArray"


class GroupUniqArray(_AggregationFunctionWithMaxSizeParam):
    function = "groupUniqArray"


class GroupArraySample(base.AggregationFunctionWithParams):
    def __init__(self, arg1, *, max_size, seed=None):
        if seed is None:
            super().__init__(arg1, params=[max_size])
        else:
            super().__init__(arg1, params=[max_size, seed])


class SumCount(base._AggregationFunction1Args):
    function = "sumCount"


class Uniq(base.AggregationFunction):
    function = "uniq"


class UniqExact(base.AggregationFunction):
    function = "uniqExact"



from clickhouse_query.functions.aggregate import (
    Any,
    ArgMax,
    ArgMin,
    Avg,
    AvgWeighted,
    Count,
    FirstValue,
    GroupArray,
    GroupArraySample,
    GroupUniqArray,
    LastValue,
    Max,
    Min,
    Sum,
    SumCount,
    TopK,
    TopKWeighted,
    Uniq,
    UniqExact,
)
from clickhouse_query.functions.arithmetic import (
    GCD,
    LCM,
    Abs,
    Divide,
    DivideDecimal,
    IntDiv,
    IntDivOrZero,
    Max2,
    Min2,
    Minus,
    Modulo,
    ModuloOrZero,
    Multiply,
    MultiplyDecimal,
    Negate,
    Plus,
    PositiveModulo,
)
from clickhouse_query.functions.base import (
    AggregationFunction,
    AggregationFunctionWithParams,
    Distinct,
    Function,
    Lambda,
)
from clickhouse_query.functions.comparison import (
    Equals,
    Greater,
    GreaterOrEquals,
    Less,
    LessOrEquals,
    NotEquals,
)
from clickhouse_query.functions.date import (
    ToDayOfWeek,
    ToDayOfYear,
    ToHour,
    ToMinute,
    ToMonth,
    ToQuarter,
    ToSecond,
    ToTime,
    ToWeek,
    ToYear,
)
from clickhouse_query.functions.in_ import In
from clickhouse_query.functions.logical import And, Not, Or, Xor
from clickhouse_query.functions.nullable import IsNull
from clickhouse_query.functions.searching_in_strings import ILike, Like
from clickhouse_query.functions.strings import Concat

__all__ = [
    # base
    "Function",
    "AggregationFunction",
    "AggregationFunctionWithParams",
    "Lambda",
    "Distinct",
    # arithmetic
    "Plus",
    "Minus",
    "Multiply",
    "Divide",
    "IntDiv",
    "IntDivOrZero",
    "Modulo",
    "ModuloOrZero",
    "PositiveModulo",
    "Negate",
    "Abs",
    "GCD",
    "LCM",
    "Max2",
    "Min2",
    "MultiplyDecimal",
    "DivideDecimal",
    # comparison
    "Equals",
    "NotEquals",
    "Less",
    "LessOrEquals",
    "Greater",
    "GreaterOrEquals",
    # date
    "ToYear",
    "ToQuarter",
    "ToMonth",
    "ToDayOfYear",
    "ToDayOfWeek",
    "ToHour",
    "ToMinute",
    "ToSecond",
    "ToTime",
    "ToWeek",
    # in_
    "In",
    # logical
    "And",
    "Or",
    "Not",
    "Xor",
    # nullable
    "IsNull",
    # searching_in_strings
    "Like",
    "ILike",
    # strings
    "Concat",
    # aggregate
    "Count",
    "Min",
    "Max",
    "Sum",
    "Avg",
    "Any",
    "FirstValue",
    "LastValue",
    "ArgMin",
    "ArgMax",
    "AvgWeighted",
    "TopK",
    "TopKWeighted",
    "GroupArray",
    "GroupUniqArray",
    "GroupArraySample",
    "SumCount",
    "Uniq",
    "UniqExact",
]



from clickhouse_query.functions import base


class ToYear(base._Function1Args):
    function = "toYear"


class ToQuarter(base._Function1Args):
    function = "toQuarter"


class ToMonth(base._Function1Args):
    function = "toMonth"


class ToDayOfYear(base._Function1Args):
    function = "toDayOfYear"


class ToDayOfWeek(base._Function1Args):
    function = "toDayOfWeek"


class ToHour(base._Function1Args):
    function = "toHour"


class ToMinute(base._Function1Args):
    function = "toMinute"


class ToSecond(base._Function1Args):
    function = "toSecond"


class ToTime(base._Function1Args):
    function = "toTime"


class ToWeek(base.Function):
    function = "toWeek"

    def __init__(self, arg, mode=None):
        if mode is None:
            super().__init__(arg)
        else:
            super().__init__(arg, mode)



from clickhouse_query.functions import base


class Plus(base._Function2Args):
    function = "plus"


class Minus(base._Function2Args):
    function = "minus"


class Multiply(base._Function2Args):
    function = "multiply"


class Divide(base._Function2Args):
    function = "divide"


class IntDiv(base._Function2Args):
    function = "intDiv"


class IntDivOrZero(base._Function2Args):
    function = "intDivOrZero"


class Modulo(base._Function2Args):
    function = "modulo"


class ModuloOrZero(base._Function2Args):
    function = "moduloOrZero"


class PositiveModulo(base._Function2Args):
    function = "positiveModulo"


class Negate(base._Function2Args):
    function = "negate"


class Abs(base._Function2Args):
    function = "abs"


class GCD(base._Function2Args):
    function = "gcd"


class LCM(base._Function2Args):
    function = "lcm"


class Max2(base._Function2Args):
    function = "max2"


class Min2(base._Function2Args):
    function = "min2"


class MultiplyDecimal(base.Function):
    function = "multiplyDecimal"

    def __init__(self, arg1, arg2, result_scale=None):
        if result_scale is not None:
            super().__init__(arg1, arg2, result_scale)
        else:
            super().__init__(arg1, arg2)


class DivideDecimal(base.Function):
    function = "divideDecimal"

    def __init__(self, arg1, arg2, result_scale=None):
        if result_scale is not None:
            super().__init__(arg1, arg2, result_scale)
        else:
            super().__init__(arg1, arg2)



from clickhouse_query import mixins, utils


class Function(mixins.ASMixin, mixins.ArithmeticMixin):
    function = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __sql__(self, uid_generator):
        sql_params = {}

        function_sql, function_params = self.__clickhouse_query_function_sql__(uid_generator=uid_generator)
        sql_params.update(function_params)

        function_args_sqls_list, function_args_params = self.__clickhouse_query_function_args_sqls__(
            uid_generator=uid_generator
        )
        sql_params.update(function_args_params)

        sql = "{func}({args})".format(func=function_sql, args=", ".join(function_args_sqls_list))
        return sql, sql_params

    def __clickhouse_query_function_sql__(self, *, uid_generator):
        return self.function, {}

    def __clickhouse_query_function_args_sqls__(self, *, uid_generator):
        args_sqls_list = []
        sql_params = {}
        for item in self.args:
            sql, params = utils.get_sql(item, uid_generator=uid_generator)
            args_sqls_list.append(sql)
            sql_params.update(params)
        return args_sqls_list, sql_params


class _Function0Args(Function):
    def __init__(self):
        super().__init__()


class _Function1Args(Function):
    def __init__(self, arg):
        super().__init__(arg)


class _Function2Args(Function):
    def __init__(self, arg1, arg2):
        super().__init__(arg1, arg2)


class AggregationFunction(Function):
    combinators_order = ["If"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.combinator_dict = {}

    def __clickhouse_query_function_sql__(self, *, uid_generator):
        function_sql, function_sql_params = super().__clickhouse_query_function_sql__(
            uid_generator=uid_generator
        )
        for combinator in self.combinators_order:
            if combinator in self.combinator_dict:
                function_sql += "{combinator}".format(combinator=combinator)

        return function_sql, function_sql_params

    def __clickhouse_query_function_args_sqls__(self, *, uid_generator):
        args_sqls_list, sql_params = super().__clickhouse_query_function_args_sqls__(
            uid_generator=uid_generator
        )
        for combinator in self.combinators_order:
            if combinator in self.combinator_dict:
                combinator_function_args = self.combinator_dict[combinator].get("function_args", [])
                for arg in combinator_function_args:
                    expression = utils._get_expression(arg)
                    sql, params = utils.get_sql(expression, uid_generator=uid_generator)
                    args_sqls_list.append(sql)
                    sql_params.update(params)
        return args_sqls_list, sql_params

    def if_(self, if_):
        self.combinator_dict["If"] = {"function_args": [if_]}
        return self


class AggregationFunctionWithParams(AggregationFunction):
    def __init__(self, *args, agg_params=None):
        super().__init__(*args)
        self.agg_params = agg_params

    def __clickhouse_query_function_sql__(self, *, uid_generator):
        function_sql, function_sql_params = super().__clickhouse_query_function_sql__(
            uid_generator=uid_generator
        )
        agg_params_sql_list = []
        if self.agg_params is not None:
            for agg_param in self.agg_params:
                expression = utils._get_expression(agg_param)
                sql, params = utils.get_sql(expression, uid_generator=uid_generator)
                agg_params_sql_list.append(sql)
                function_sql_params.update(params)

        function_sql = function_sql + "({agg_params})".format(agg_params=", ".join(agg_params_sql_list))
        return function_sql, function_sql_params


class _AggregationFunction0Args(AggregationFunction):
    def __init__(self):
        super().__init__()


class _AggregationFunction1Args(AggregationFunction):
    def __init__(self, arg):
        super().__init__(arg)


class _AggregationFunction2Args(AggregationFunction):
    def __init__(self, arg1, arg2):
        super().__init__(arg1, arg2)


class Lambda(Function):
    function = "lambda"

    def __init__(self, x, expr):
        super().__init__(x, expr)


class Distinct(Function):
    function = "distinct"
